from collections import deque
from hippodamia.enums import Health
from hippodamia.enums import ViewDetails
from hippodamia.enums import Necessity
import threading


"""
[ROOT]
  ├─ [ANode]
  │    ├─ [BNode]
  │    ├─ [CNode]
  │    │    ├─ [DNode]
  │    │    │    ├─ [ENode]
  │    │    │    └─ [FNode]  
  │    │    └─ [GNode]
  │    └─ [HNode]
  └─ [INode]
       ├─ [JNode]
       ├─ [KNode]
       │    ├─ [LNode]
       │    └─ [MNode]       
       └─ [NNode]
    
  
Root-ANode - prefix_first_line =  "  ├─ "
Root-ANode - prefix_next_lines =  "  │  "
ANode-CNode - prefix_first_line = "  │    ├─ "
ANode-CNode - prefix_next_lines = "  │    │  "
CNode-DNode - prefix_first_line = "  │    │    ├─ "
CNode-DNode - prefix_next_lines = "  │    │    │  "
INode-KNode - prefix_first_line = "       ├─ "
INode-KNode - prefix_next_lines = "       │  "

TYPE_A = "  ├─ "
TYPE_B = "  │  "
TYPE_C = "  └─ "
TYPE_D = "     "
"""


class Entry:
    MAX_NAME_LEN = 15
    INDENT = 2

    HORIZONTAL = "│"
    HORIZONTAL_CONNECTOR = "├"
    VERTICAL = "─"
    VERTICAL_CONNECTOR = "┬"
    EDGE = "└"
    SPACE = " "

    TYPE_A = SPACE + SPACE + HORIZONTAL_CONNECTOR + VERTICAL + SPACE
    TYPE_B = SPACE + SPACE + HORIZONTAL + SPACE + SPACE
    TYPE_C = SPACE + SPACE + EDGE + VERTICAL + SPACE
    TYPE_D = SPACE + SPACE + SPACE + SPACE + SPACE

    visited = None
    health = None
    name = None

    def update(self):
        raise NotImplementedError

    def reset_visited(self):
        raise NotImplementedError

    def _name2str(self):
        result = self.name
        if len(result) > self.MAX_NAME_LEN:
            result = result[:12]
        result = "[{}]".format(result)
        return result


class Branch(Entry):
    children = None

    def __init__(self, name):
        Entry.__init__(self)
        self.children = {}
        self.name = name
        self.update()

    def update(self, children=None):
        if not self.visited:
            if children is None:
                children = self.children

            self.health = Health.GREEN
            for gid, leaf in children.items():
                leaf_health = leaf.update()
                self.health = max(self.health, leaf_health)

            self.visited = True
        return self.health

    def reset_visited(self):
        for gid, leaf in self.children.items():
            leaf.reset_visited()
        self.visited = False

    def add_leaf(self, leaf, breadcrumbs, children=None):
        if children is None:
            children = self.children

        if len(breadcrumbs) == 0:
            children[leaf.gid] = leaf
        else:
            branch_name = breadcrumbs.popleft()
            if branch_name is None or branch_name == "":
                branch_name = "_"

            if branch_name not in children:
                children[branch_name] = Branch(branch_name)
            children[branch_name].add_leaf(leaf, breadcrumbs)

    def pformat(self, details=ViewDetails.NONE, filters=None):
        if filters is None:
            filters = deque()
        lines = self.tree_view("", "", details, filters)
        result = "\n".join(lines)
        return  result

    @staticmethod
    def _filter_match(name, entry_filter):
        if entry_filter is not None:
            name = name.lower()
            entry_filter = entry_filter.lower()
            return name.startswith(entry_filter)
        return True

    def tree_view(self, prefix_first_line, prefix_next_lines, details, filters):
        filters = filters.copy()
        try:
            entry_filter = filters.popleft()
        except IndexError:
            entry_filter = None

        first_line = prefix_first_line + self._pprint_details(details)
        lines = [first_line]
        prefix_a = prefix_next_lines + self.TYPE_A
        prefix_b = prefix_next_lines + self.TYPE_B
        prefix_c = prefix_next_lines + self.TYPE_C
        prefix_d = prefix_next_lines + self.TYPE_D

        counter = 0
        for name, entry in self.children.items():
            counter += 1
            if self._filter_match(name, entry_filter):
                if counter == len(self.children):
                    # last entry
                    prefix_first = prefix_c
                    prefix_next = prefix_d
                else:
                    prefix_first = prefix_a
                    prefix_next = prefix_b

                if type(entry) is Leaf:
                    line = prefix_first + entry.pformat(details)
                    lines.append(line)
                elif type(entry) is Branch:
                    sub_lines = entry.tree_view(prefix_first, prefix_next, details, filters)
                    lines += sub_lines
                else:
                    raise TypeError

        return lines

    def _pprint_details(self, details):
        result = [self._name2str()]
        if details & ViewDetails.HEALTH == ViewDetails.HEALTH:
            result.append(self.health.name)
        if details & ViewDetails.STATE == ViewDetails.STATE:
            pass
        if details & ViewDetails.GID == ViewDetails.GID:
            pass
        result = " ".join(result)
        return result


class Root(Branch):
    def __init__(self):
        Branch.__init__(self, name="__root__")

    def build_tree(self, leafs):
        new_children = {}

        for gid, leaf in leafs.items():
            breadcrumbs = leaf.breadcrumbs.copy()
            self.add_leaf(leaf, breadcrumbs, new_children)

        self.children = new_children
        self.update()


class Leaf(Entry):
    gid = None
    shadow = None
    breadcrumbs = None

    def __init__(self, shadow):
        Entry.__init__(self)
        self.shadow = shadow
        self.gid = self.shadow.properties.gid
        self.breadcrumbs = deque()
        self.update()

    def update(self):
        self.visited = True
        self.health = self.shadow.properties.health
        self.name = self.shadow.properties.name
        self.breadcrumbs = deque([self.shadow.properties.location, self.shadow.properties.room,
                                  self.shadow.properties.device])
        result = self.health
        if self.shadow.properties.necessity != Necessity.REQUIRED:
            result = min(result, Health.YELLOW)

        return result

    def reset_visited(self):
        self.visited = False

    def pformat(self, details=ViewDetails.NONE):
        result = [self._name2str()]

        if details & ViewDetails.HEALTH == ViewDetails.HEALTH:
            result.append(self.health.name)
        if details & ViewDetails.STATE == ViewDetails.STATE:
            result.append(self.shadow.get_state_id().name)
        if details & ViewDetails.GID == ViewDetails.GID:
            result.append("gid:")
            result.append(self.gid)

        result = " ".join(result)
        return result


class HierarchicalView:
    _logger = None
    _agentshadows = None
    _enforcement = None

    _tree = None
    _leafs = None

    _updateavailable = None
    _update_observer_thread = None
    _stop_observer = None

    def __init__(self, enforcement, agentshadows, updateavailable, logger):
        self._enforcement = enforcement
        self._logger = logger
        self._agentshadows = agentshadows
        self._updateavailable = updateavailable

        self._tree = Root()
        self._leafs = {}

        self._update_observer_thread = threading.Thread(target=self._update_observer)
        self._stop_observer = False

    def _process_update(self):
        self._tree.reset_visited()
        rebuild = self._add_new_shadows()
        rebuild = rebuild or self._purge_unvisited_leafs()
        if rebuild:
            self._tree.build_tree(self._leafs)
        self._tree.update()

    def pformat(self, details=ViewDetails.HEALTH, filters=None):
        if filters is None:
            filters = deque([None, None, None, None])
        else:
            filters = deque([filters["LOCATION"], filters["ROOM"], filters["DEVICE"], filters["SERVICE"]])
        return self._tree.pformat(details, filters)

    def _purge_unvisited_leafs(self):
        gids = []
        for gid, leaf in self._leafs.items():
            if not leaf.visited:
                gids.append(gid)
        rebuild = len(gids) > 0

        for gid in gids:
            del self._leafs[gid]

        return rebuild

    def _add_new_shadows(self):
        rebuild_tree = False
        for gid, shadow in self._agentshadows.items():
            if gid not in self._leafs:
                leaf = Leaf(shadow)
                self._leafs[gid] = leaf
                rebuild_tree = True
            else:
                self._leafs[gid].visited = True

        return rebuild_tree

    def _update_observer(self):
        while not self._stop_observer:
            self._updateavailable.wait()
            if not self._stop_observer:
                self._process_update()
            self._updateavailable.clear()

    def start(self):
        self._stop_observer = False
        self._update_observer_thread.start()
        self._updateavailable.set()

    def stop(self):
        self._stop_observer = True
        self._updateavailable.set()
        self._update_observer_thread.join()



from hippodamia.enums import Health


class Properties:
    _logger = None

    _gid = None
    _necessity = None
    _type = None
    _name = None
    _location = None
    _room = None
    _device = None

    uuid = None
    onboarding_topic = None
    protocol_version = None
    version = None
    description = None
    host_name = None
    node_id = None
    ips = None
    config_hash = None

    state_health = None
    log_health = None

    def __init__(self, logger):
        self._logger = logger

    def export_dict(self):
        try:
            health = self.health.name
        except AttributeError:
            health = None
        try:
            necessity = self.necessity.name
        except AttributeError:
            necessity = None

        result = {
            "gid": self.gid,
            "health": health,
            "necessity": necessity,
            "uuid": self.uuid,
            "onboarding_topic": self.onboarding_topic,
            "protocol_version": self.protocol_version,
            "type": self.type,
            "version": self.version,
            "name": self.name,
            "location": self.location,
            "room": self.room,
            "device": self.device,
            "description": self.description,
            "host_name": self.host_name,
            "node_id": self.node_id,
            "ips": self.ips,
            "config_hash": self.config_hash
        }
        return result

    def _set_necessity(self, necessity):
        if self._necessity is None:
            self._necessity = necessity
        elif self._necessity != necessity:
            msg = "set_necessity() is called with different value ({}) than the existing one ({})".format(necessity, self._necessity)
            self._logger.warning(msg)
            raise ValueError(msg)

    def _get_necessity(self):
        return self._necessity

    necessity = property(_get_necessity, _set_necessity)

    def _set_gid(self, gid):
        if self._gid is None:
            self._gid = str(gid)
        elif self._gid != str(gid):
            msg = "set_gid() is called with different value ({}) than the existing one ({})".format(gid, self._gid)
            self._logger.warning(msg)
            raise ValueError(msg)

    def _get_gid(self):
        return self._gid

    gid = property(_get_gid, _set_gid)
    
    def _set_type(self, type):
        if self._type is None:
            self._type = type
        elif self._type != type:
            msg = "set_type() is called with different value ({}) than the existing one ({})".format(type, self._type)
            self._logger.warning(msg)
            raise ValueError(msg)

    def _get_type(self):
        return self._type

    type = property(_get_type, _set_type)

    def _set_name(self, name):
        if self._name is None:
            self._name = name
        elif self._name != name:
            msg = "set_name() is called with different value ({}) than the existing one ({})".format(name, self._name)
            self._logger.warning(msg)
            raise ValueError(msg)

    def _get_name(self):
        return self._name

    name = property(_get_name, _set_name)
    
    def _set_location(self, location):
        if self._location is None:
            self._location = location
        elif self._location != location:
            msg = "set_location() is called with different value ({}) than the existing one ({})".format(location, self._location)
            self._logger.warning(msg)
            raise ValueError(msg)

    def _get_location(self):
        return self._location

    location = property(_get_location, _set_location)
    
    def _set_room(self, room):
        if self._room is None:
            self._room = room
        elif self._room != room:
            msg = "set_room() is called with different value ({}) than the existing one ({})".format(room, self._room)
            self._logger.warning(msg)
            raise ValueError(msg)

    def _get_room(self):
        return self._room

    room = property(_get_room, _set_room)

    def _set_device(self, device):
        if self._device is None:
            self._device = device
        elif self._device != device:
            msg = "set_device() is called with different value ({}) than the existing one ({})".format(device, self._device)
            self._logger.warning(msg)
            raise ValueError(msg)

    def _get_device(self):
        return self._device

    device = property(_get_device, _set_device)

    def _get_health(self):
        try:
            result = max(self.log_health, self.state_health)
        except TypeError:
            try:
                result = max(self.log_health)
            except TypeError:
                try:
                    result = max(self.state_health)
                except TypeError:
                    result = Health.RED
        return result

    health = property(fget=_get_health)




import json
import threading
import datetime
import pelops.mylogger


class Heartbeat:
    interval = None
    topic = None
    _logger = None
    _mqtt_client = None
    _stop_timer = None

    counter = 0
    last_timestamp = None

    def __init__(self, topic, interval, mqtt_client, logger):
        self.topic = topic
        self.interval = interval
        self._mqtt_client = mqtt_client
        self._logger = pelops.mylogger.get_child(logger, __class__.__name__)
        self._stop_timer = threading.Event()
        self._timer_thread = threading.Thread(target=self._timer)
        self._logger.info("__init__ done")

    def _publish(self):
        """
        {
            "heartbeat": "heartbeat",
            "timestamp": "1985-04-12T23:20:50.520Z"
        }
        """
        message = {
            "heartbeat": "heartbeat",
            "timestamp": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")  # "2009-11-10T23:00:00.0Z"
        }
        self.counter += 1
        self.last_timestamp = message["timestamp"]
        self._mqtt_client.publish(self.topic, json.dumps(message))
        self._logger.info("published heartbeat ({})".format(self.counter))

    def _timer(self):
        while True:
            self._stop_timer.wait(self.interval)
            if self._stop_timer.isSet():
                break
            self._publish()

    def start(self):
        self._publish()
        self._timer_thread.start()
        self._logger.info("start finished")

    def stop(self):
        self._stop_timer.set()
        self._timer_thread.join()
        self._logger.info("stop finished")



import datetime
from pelops.mylogger import get_log_level
from logging import NOTSET
import collections
import threading


class ShadowLogging:
    class Entry:
        log = None
        counter = None
        time = None

        def __init__(self, maxlen):
            self.log = collections.deque(maxlen=maxlen)
            self.counter = 0

        def add(self, message):
            self.log.appendleft(message)
            self.counter += 1
            self.time = datetime.datetime.now()

        def stats(self):
            output = "received: {}, stored: {}, @{}".format(self.counter, len(self.log), self.time)
            return output

    flat_list = None
    level_dict = None
    max_level = None
    set_health = None

    def __init__(self, maxlen=50, set_health=None):
        self.set_health = set_health
        self.reset_max_level()
        self.flat_list = ShadowLogging.Entry(maxlen)
        self.level_dict = collections.defaultdict(lambda: ShadowLogging.Entry(maxlen))

    def add(self, message):
        self.flat_list.add(message)
        level = get_log_level(message["level"])
        self.level_dict[level].add(message)
        self.max_level = max(self.max_level, level)
        self.set_health(self.max_level)

    def reset_max_level(self):
        self.max_level = NOTSET
        self.set_health(self.max_level)


class ShadowData:
    gid = None

    properties = None
    properties_time = None

    request = {}
    request_counter = 0
    request_time = None

    config = {}
    config_counter = 0
    config_time = None

    runtime = {}
    runtime_counter = 0
    runtime_time = None

    ping = {}
    ping_counter = 0
    ping_time = None

    end = {}
    end_counter = 0
    end_time = None

    logger = None

    def __init__(self, gid, set_log_health):
        self.gid = str(gid)
        self.logger = ShadowLogging(50, set_log_health)

    def process_properties(self, properties):
        self.properties = properties.export_dict()
        self.properties_time = datetime.datetime.now()

    def process_ping(self, message):
        self.ping = message
        self.ping_counter += 1
        self.ping_time = datetime.datetime.now()

    def process_config(self, message):
        self.config = message
        self.config_counter += 1
        self.config_time = datetime.datetime.now()

    def process_runtime(self, message):
        self.runtime = message
        self.runtime_counter += 1
        self.runtime_time = datetime.datetime.now()

    def process_end(self, message):
        self.end = message
        self.end_counter += 1
        self.end_time = datetime.datetime.now()

    def process_logger(self, message):
        self.logger.add(message)

    def _get_logger_time(self):
        return self.logger.flat_list.time

    logger_time = property(fget=_get_logger_time)

    def _get_logger_counter(self):
        return self.logger.flat_list.counter

    logger_counter = property(fget=_get_logger_counter)

    def process_request(self, message):
        self.request = message
        self.request_counter += 1
        self.request_time = datetime.datetime.now()

    def get_counter_dict(self):
        result = {
            "request": self.request_counter,
            "ping": self.ping_counter,
            "runtime": self.runtime_counter,
            "config": self.config_counter,
            "logger": self.logger_counter,
            "end": self.end_counter
        }
        return result

    def get_stats(self):
        return "request {}, ping {}, runtime {}, config {}, logger {}, end {}"\
            .format(self.request_counter, self.ping_counter, self.runtime_counter,
                    self.config_counter, self.logger_counter, self.end_counter)



import datetime
from hippodamia.enums import Enforcement
import json
import pelops.mylogger


class Onboarding:
    _logger = None
    _mqtt_client = None
    _agentshadows = None
    _asfactory = None

    enforcement = None
    protocol_version = None
    topic_onboarding_request = None
    counter = 0
    gid_list = set()
    last_timestamp = None
    last_message = None

    def __init__(self, protocol_version, topic_onboarding_request, enforcement, agentshadows, asfactory, mqtt_client, logger):
        self.protocol_version = protocol_version
        self._logger = pelops.mylogger.get_child(logger, __class__.__name__)
        self._mqtt_client = mqtt_client
        self.enforcement = enforcement
        self._agentshadows = agentshadows
        self._asfactory = asfactory
        self.topic_onboarding_request = topic_onboarding_request
        self._logger.info("__init__ done")

    def start(self):
        self._mqtt_client.subscribe(self.topic_onboarding_request, self._handler_onboarding_request)
        self._logger.info("subscribed to {}".format(self.topic_onboarding_request))

    def stop(self):
        self._mqtt_client.unsubscribe(self.topic_onboarding_request, self._handler_onboarding_request)
        self._logger.info("unsubscribed from {}".format(self.topic_onboarding_request))

    def _handler_onboarding_request(self, message):
        """
        {
            "uuid": "550e8400-e29b-11d4-a716-446655440000",
            "onboarding-topic": "/hippodamia/550e8400-e29b-11d4-a716-446655440000",
            "protocol-version": 1,
            "timestamp": "1985-04-12T23:20:50.520Z",
            "identifier": {
                "gid": 1,
                "type": "copreus",
                "name": "display-driver",
                "location": "flat",
                "room": "living room",
                "device": "thermostat",
                "decription": "lorem ipsum",
                "host-name": "rpi",
                "node-id": "00-07-E9-AB-CD-EF",
                "ips": [
                    "192.168.0.1",
                    "10.0.1.2",
                    "2001:0db8:85a3:08d3:1319:8a2e:0370:7344"
                ],
                "config-hash": "cf23df2207d99a74fbe169e3eba035e633b65d94"
            }
        }
        """
        message = message.decode("utf8")
        message = json.loads(message)
        self.last_message = message
        self._logger.info("_handler_onboarding_request - received request")
        self._logger.debug("_handler_onboarding_request - message: {}".format(message))

        self.counter += 1
        self.last_timestamp = datetime.datetime.now()

        if message["protocol-version"] != self.protocol_version:
            message = "handler_onboarding_request - expected protocol version {}, message: {}"\
                .format(self.protocol_version, message)
            self._logger.error(message)
        else:
            identifier = message["identifier"]
            self._logger.debug("_handler_onboarding_request - identifier: {}".format(identifier))
            gid = str(identifier["gid"])
            self.gid_list.add(gid)
            self._logger.debug("_handler_onboarding_request - gid: {}".format(gid))

            shadow = None
            if gid not in self._agentshadows:
                self._logger.debug("_handler_onboarding_request - unknown gid")
                if self.enforcement == Enforcement.NONE:
                    self._logger.debug("_handler_onboarding_request - Enforcement.NONE")
                    shadow = self._asfactory.new_agentshadow(gid=gid)
                    self._agentshadows[gid] = shadow
                    shadow.start()
                elif self.enforcement == Enforcement.IGNORE:
                    self._logger.debug("_handler_onboarding_request - Enforcement.IGNORE")
                    return
                elif self.enforcement == Enforcement.STRICT:
                    self._logger.debug("_handler_onboarding_request - Enforcement.STRICT")
                    self._logger.error("handler_onboarding_request - required is set to strict and incoming is gid "
                                       "not pre-configured. '{}'".format(message))
                    return
                else:
                    message = "handler_onboarding_request - dont know how to handler required '{}'"\
                        .format(self.enforcement)
                    self._logger.error(message)
                    raise ValueError(message)
            else:
                self._logger.debug("_handler_onboarding_request - existing gid {}".format(gid))
                shadow = self._agentshadows[gid]

            self._logger.debug("_handler_onboarding_request - {}".format(shadow))
            shadow.process_onboarding_request(message)



from enum import Enum
from enum import IntEnum


class Enforcement(Enum):
    NONE = 1
    IGNORE = 2
    STRICT = 3

    @classmethod
    def get_enum(cls, text):
        text = text.upper()
        if text == "NONE":
            return Enforcement.NONE
        elif text == "IGNORE":
            return Enforcement.IGNORE
        elif text == "STRICT":
            return Enforcement.STRICT
        else:
            raise ValueError("dont know how to convert '{}' to enum Enforcement.".format(text))


class Necessity(Enum):
    OPTIONAL = 1  # predefined in config, but if inactive/stopped, the combined group health is still green
    REQUIRED = 2  # predefiend in config, if inactive/stopped, the combined group health is red
    SPONTANEOUS = 3  # not predefined. if configured, occurence of a spontaneous agent onboarding will result in a red group health

    @classmethod
    def get_enum(cls, text):
        text = text.upper()
        if text == "OPTIONAL":
            return Necessity.OPTIONAL
        elif text == "REQUIRED":
            return Necessity.REQUIRED
        elif text == "SPONTANEOUS":
            return Necessity.SPONTANEOUS
        else:
            raise ValueError("dont know how to convert '{}' to Necessity".format(text))


class StatType(Enum):
    ALL = 1
    GENERAL = 2
    PROPERTIES = 3
    PING = 4
    RUNTIME = 5
    CONFIG = 6
    LOGGER = 7
    REQUEST = 8

    @classmethod
    def get_enum(cls, text):
        text = str(text).upper()
        for name, member in StatType.__members__.items():
            if text == name:
                return member
        raise ValueError("StatType: dont know how to convert '{}' to enum".format(text))

    @classmethod
    def get_members(cls):
        return StatType.__members__.keys()


class Health(IntEnum):
    GREEN = 0
    YELLOW = 1
    RED = 2


class ViewDetails(IntEnum):
    NONE = 0
    HEALTH = 1
    STATE = 2
    GID = 4
    FULL = NONE+HEALTH+STATE+GID



def get_schema():
    schema = {
        "monitoringservice": {
            "description": "Hippodamia observes the state of all registered microservices (aka watch dog).",
            "type": "object",
            "properties": {
                "topics": _topics(),
                "service-timings": _service_timings(),
                "agent-timings": _agent_timings(),
                "expected-services": _expected_services(),
                "mongodb": _mongodb(),
                "webpage": _webpage()
            },
            "required": ["topics", "agent-timings", "service-timings", "expected-services", "mongodb", "webpage"],
            "additionalProperties": False
        }
    }
    return schema


def _topics():
    schema = {
        "type": "object",
        "description": "all topics for the monitoring services",
        "properties": {
            "onboarding-request": {
                "type": "string",
                "example": "/test/hippodamia/onboarding/request",
                "description": "hippodamia listens on this topic for new onboarding request"
            },
            "activity-base": {
                "type": "string",
                "example": "/test/hippodamia/incoming/",
                "description": "this topic will be used as base for each registered service. e.g .../incoming/1/ping"
            },
            "command-end": {
                "type": "string",
                "example": "/test/hippodamia/command/end",
                "description": "end command messages are published to this topic"
            },
            "command-reonboarding": {
                "type": "string",
                "example": "/test/hippodamia/command/reonboarding",
                "description": "reonboarding request are published to this topic"
            },
            "command-ping-on-request": {
                "type": "string",
                "example": "/test/hippodamia/command/ping_on_request",
                "description": "spontaneous ping messages are requested via this topic"
            },
            "command-config-on-request": {
                "type": "string",
                "example": "/test/hippodamia/command/command/config_on_request",
                "description": "spontaneous config messages are requested via this topic"
            },
            "command-runtime-on-request": {
                "type": "string",
                "example": "/test/hippodamia/command/command/runtime_on_request",
                "description": "spontaneous runtime messages are requested via this topic"
            },
            "heartbeat": {
                "type": "string",
                "example": "/test/hippodamia/heartbeat",
                "description": "heartbeat is published via this topic"
            }
        },
        "required": ["onboarding-request", "activity-base", "command-end", "command-reonboarding", "command-ping-on-request",
                     "command-config-on-request", "command-runtime-on-request", "heartbeat"],
        "additionalProperties": False
    }
    return schema


def _agent_timings():
    schema = {
                "type": "object",
                "description": "intervals in seconds for different tasks of the monitoring agent - transmitted in the "
                               "onboarding response message",
                "properties": {
                    "send-ping": {
                        "description": "interval in seconds for the scheduler to send a ping message to the "
                                       "monitoring service. 0 deactivates the scheduler. example: 60 seconds.",
                        "type": "integer"
                    },
                    "send-runtime": {
                        "description": "interval in seconds for the scheduler to send a runtime message to the "
                                       "monitoring service. 0 deactivates the scheduler. example: 500 seconds.",
                        "type": "integer"
                    },
                    "send-config": {
                        "description": "interval in seconds for the scheduler to send a config message to the "
                                       "monitoring service. 0 deactivates the scheduler. example: 3600 seconds.",
                        "type": "integer"
                    },
                    "expect-heartbeat": {
                        "description": "maximum interval in seconds until a heartbeat message from the monitoring"
                                       "service should be received. If violated, the agent start the re-onboarding"
                                       "procedures. 0 deactivates this feature. example: 500 seconds.",
                        "type": "integer"
                    },
                },
                "required": ["send-ping", "send-runtime", "send-config", "expect-heartbeat"]
            }
    return schema


def _service_timings():
    schema = {
        "type": "object",
        "description": "all timings for the monitoring service",
        "properties": {
            "onboarding_timeout": {
                "type": "integer",
                "example": "60",
                "description": "seconds * 2 to be waited until an onboarding request should have been processed and a response been recevied"
            },
            "wait_for_message_timeout": {
                "type": "integer",
                "example": "180",
                "description": "seconds - if no message has been received for this period, the service is considered to be inactive"
            },
            "deactivation_timeout": {
                "type": "integer",
                "example": "360",
                "description": "seconds - an inactive service becomes a stopped service after this time span"
            },
            "heartbeat_interval": {
                "type": "integer",
                "example": "60",
                "description": "seconds - send heartbeat to all connected services"
            },
            "archivation_timeout": {
                "type": "integer",
                "example": "86400",
                "description": "seconds - a stopped service is archived after this time span."
            }
        },
        "required": ["onboarding_timeout", "wait_for_message_timeout", "deactivation_timeout",
                     "heartbeat_interval", "archivation_timeout"],
        "additionalProperties": False
    }
    return schema


def _expected_services():
    schema = {
        "type": "object",
        "description": "pre-configure which services are expected to be onboarded",
        "properties": {
            "enforcement": {
                "type": "string",
                "example": "none",
                "description": "spontaneously appearing services are either [none]-accepted, [ignore]-ignored, "
                               "[strict]-leads to a error log entry",
                "enum": ["none", "ignore", "strict"]
            },
            "services": {
                "type": "array",
                "description": "list of expected services",
                "items": {
                    "type": "object",
                    "properties": {
                        "gid": {
                            "type": ["string", "integer"],
                            "example": "1",
                            "description": "gid of the service"
                        },
                        "name": {
                            "type": "string",
                            "example": "Thermostat GUI",
                            "description": "name of the service (optional)"
                        },
                        "location": {
                            "type": "string",
                            "example": "vienna",
                            "description": "location of the service (optional)"
                        },
                        "room": {
                            "type": "string",
                            "example": "living room",
                            "description": "room of the service (optional)"
                        },
                        "device": {
                            "type": "string",
                            "example": "thermostat",
                            "description": "device the service is running on (optional)"
                        },
                        "necessity": {
                            "type": "string",
                            "example": "required",
                            "description": "configure if the expected service is [required] or [optional]",
                            "enum": ["required", "optional"]
                        }
                    },
                    "required": ["gid", "necessity"],
                    "additionalProperties": False
                }
            }
        },
        "required": ["enforcement", "services"],
        "additionalProperties": False
    }
    return schema


def _mongodb():
    schema = {
        "type": "object",
        "description": "persistance database",
        "properties": {
            "credentials-file": {
                "type": "string",
                "example": "~/credentials.yaml",
                "description": "credentials for mongodb"
            },
            "mongodb-address": {
                "type": "string",
                "example": "localhost",
                "description": "host name"
            },
            "mongodb-port": {
                "type": "integer",
                "example": 21017,
                "description": "port"
            },
            "database": {
                "type": "string",
                "example": "hippodamia",
                "description": "database name"
            },
            "mongodb-user": {
                "description": "User name for mongodb",
                "type": "string"
            },
            "mongodb-password": {
                "description": "Password for mongodb",
                "type": "string"
            },
            "log-level": {
                "type": "string",
                "example": "DEBUG",
                "description": "log level for mongodb client related messages",
                "enum": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
            }
        },
        "required": ["mongodb-address", "mongodb-port", "database", "log-level"],
        "additionalProperties": False
    }
    return schema


def _webpage():
    schema = {
        "type": "object",
        "description": "web page to view the state of the services",
        "properties": {
            "port": {
                "type": "integer",
                "example": "6060",
                "description": "web server port"
            },
            "webpage-user": {
                "description": "User name for webpage",
                "type": "string"
            },
            "webpage-password": {
                "description": "Password for webpage",
                "type": "string"
            },
            "credentials-file": {
                "type": "string",
                "example": "~/credentials.yaml",
                "description": "if credentials are set, website is password protected"
            },
            "log-level": {
                "type": "string",
                "example": "DEBUG",
                "description": "log level for mongodb client related messages",
                "enum": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
            }
        },
        "required": ["port", "log-level"],
        "additionalProperties": False
    }
    return schema



from hippodamia.agentshadow.states.amonitoringstate import AMonitoringState
from hippodamia.enums import Health


class Onboarded(AMonitoringState):
    func_activate_recv_messages = None

    def __init__(self, update_available, set_health, reset_max_level, logger):
        AMonitoringState.__init__(self, update_available, set_health, Health.YELLOW, reset_max_level,
                                  logger, __class__.__name__)

    def _on_entry(self):
        return None

    def _on_exit(self):
        pass



import threading
import json

from pelops.abstractmicroservice import AbstractMicroservice
import hippodamia
from hippodamia.agentshadow.agentshadowfactory import AgentShadowFactory
from hippodamia.mymongoclient import MyMongoClient
from hippodamia.heartbeat import Heartbeat
from hippodamia.commands import CommandFactory
from hippodamia.schema.monitoringservice import get_schema
from hippodamia.hierarchicalview import HierarchicalView
from hippodamia.enums import Enforcement
from hippodamia.onboarding import Onboarding


class Monitoringservice(AbstractMicroservice):
    _version = hippodamia.version

    _protocol_version = 1

    _agentshadows = None
    _asfactory = None
    _update_available = None
    _mongo_client = None
    _heartbeat = None
    _hierarchical_view = None
    _onboarding = None

    _enforcement = None

    _topic_onboarding_request = None
    _topic_cmd_ping = None
    _topic_cmd_runtime = None
    _topic_cmd_config = None
    _topic_cmd_reonboarding = None
    _topic_cmd_end = None

    def __init__(self, config, mqtt_client=None, logger=None, stdout_log_level=None, no_gui=None):
        """
        Constructor - creates the services and the tasks

        :param config: config yaml structure
        :param mqtt_client: mqtt client instance
        :param logger: logger instance
        :param no_gui: if False create and control a ui instance
        :param stdout_log_level: if set, a logging handler with target sys.stdout will be added
        """
        AbstractMicroservice.__init__(self, config, "monitoringservice", mqtt_client, logger,
                                      logger_name=__class__.__name__, stdout_log_level=stdout_log_level, no_gui=no_gui)
        self._mongo_client = MyMongoClient(self._config["mongodb"], self._logger)
        self._heartbeat = Heartbeat(self._config["topics"]["heartbeat"],
                                    self._config["service-timings"]["heartbeat_interval"],
                                    self._mqtt_client, self._logger)

        self._topic_onboarding_request = self._config["topics"]["onboarding-request"]
        self._topic_cmd_ping = self._config["topics"]["command-ping-on-request"]
        self._topic_cmd_runtime = self._config["topics"]["command-runtime-on-request"]
        self._topic_cmd_config = self._config["topics"]["command-config-on-request"]
        self._topic_cmd_reonboarding = self._config["topics"]["command-reonboarding"]
        self._topic_cmd_end = self._config["topics"]["command-end"]

        self._update_available = threading.Event()
        self._update_available.clear()

        self._asfactory = AgentShadowFactory(self._config["topics"], self._config["service-timings"],
                                             self._config["agent-timings"], self._update_available,
                                             self._mongo_client, self._mqtt_client, self._logger)
        self._agentshadows = self._asfactory.create_agentshadows(self._config["expected-services"]["services"])

        self._enforcement = Enforcement.get_enum(self._config["expected-services"]["enforcement"])
        self._hierarchical_view = HierarchicalView(self._enforcement, self._agentshadows, self._update_available,
                                                   self._logger)

        self._onboarding = Onboarding(self._protocol_version, self._topic_onboarding_request, self._enforcement,
                                      self._agentshadows, self._asfactory, self._mqtt_client, self._logger)

        factory = CommandFactory(self._version, self._config, self._agentshadows, self._heartbeat,
                                 self._hierarchical_view, self._onboarding, self.send_cmd_ping, self.send_cmd_runtime,
                                 self.send_cmd_config, self.send_cmd_reonboarding, self.send_cmd_end, self.stop,
                                 self._logger)
        factory.add_commands(self._add_ui_command)

    def _start(self):
        self._mongo_client.start()
        self._heartbeat.start()
        for k, v in self._agentshadows.items():
            v.start()
        self._hierarchical_view.start()
        self._onboarding.start()
        self.send_cmd_reonboarding(gid=None)  # send reonboarding request to all

    def _stop(self):
        self._heartbeat.stop()
        self._onboarding.stop()
        self.send_cmd_end(gid=None)  # inform everyone that the service is going down
        self._hierarchical_view.stop()
        for k, v in self._agentshadows.items():
            print("stopping {}".format(k))
            v.stop()
            import time
            time.sleep(1)
        self._mongo_client.stop()

    @classmethod
    def _get_description(cls):
        return "Hippodamia observes the state of all registered microservices (aka watch dog)."

    @classmethod
    def _get_schema(cls):
        return get_schema()

    def runtime_information(self):
        info = {}
        for gid, shadow in self._agentshadows.items():
            entry = {
                "gid": gid,
                "state": shadow.get_state_id().name,
                "health": shadow.properties.health.name,
                "necessity": shadow.properties.necessity.name,
                "properties": shadow.properties.name
            }
            info[gid] = entry
        return info

    def config_information(self):
        return {}

    @staticmethod
    def _render_message(request, gid):
        """
        {
            "request": "request",
            "gid": 1
        }
        """
        message = {
            "request": request
        }
        if gid is not None:
            message["gid"] = gid
        return json.dumps(message)

    def send_cmd_ping(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_ping: send request to all monitored services")
        else:
            self._logger.info("send_cmd_ping: send request to gid: {}".format(gid))
        message = self._render_message("ping", gid)
        self._mqtt_client.publish(self._topic_cmd_ping, message)

    def send_cmd_runtime(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_runtime: send request to all monitored services")
        else:
            self._logger.info("send_cmd_runtime: send request to gid: {}".format(gid))
        message = self._render_message("runtime", gid)
        self._mqtt_client.publish(self._topic_cmd_runtime, message)

    def send_cmd_config(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_config: send request to all monitored services")
        else:
            self._logger.info("send_cmd_config: send request to gid: {}".format(gid))
        message = self._render_message("config", gid)
        self._mqtt_client.publish(self._topic_cmd_config, message)

    def send_cmd_reonboarding(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_reonboarding: send request to all monitored services")
        else:
            self._logger.info("send_cmd_reonboarding: send request to gid: {}".format(gid))
        message = self._render_message("reonboarding", gid)
        self._mqtt_client.publish(self._topic_cmd_reonboarding, message)

    def send_cmd_end(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_end: send request to all monitored services")
        else:
            self._logger.info("send_cmd_end: send request to gid: {}".format(gid))
        message = self._render_message("end", gid)
        self._mqtt_client.publish(self._topic_cmd_end, message)


def standalone():
    Monitoringservice.standalone()


if __name__ == "__main__":
    Monitoringservice.standalone()



from hippodamia.agentshadow.agentshadow import AgentShadow


class AgentShadowFactory:
    _topics = None
    _service_timings = None
    _agent_timings = None
    _update_available = None
    _mongo_client = None
    _mqtt_client = None
    _logger = None

    def __init__(self, topics, service_timings, agent_timings, update_available, mongo_client, mqtt_client, logger):
        self._topics = topics
        self._service_timings = service_timings
        self._agent_timings = agent_timings
        self._update_available = update_available
        self._mongo_client = mongo_client
        self._mqtt_client = mqtt_client
        self._logger = logger

    def create_agentshadows(self, preload_config_list):
        shadows = {}

        for preload_config in preload_config_list:
            gid = str(preload_config["gid"])
            shadow = self.new_agentshadow(preload_config=preload_config)
            shadows[gid] = shadow

        return shadows

    def new_agentshadow(self, preload_config=None, gid=None):
        if preload_config is None and gid is None:
            raise ValueError("AgentShadowFactory.create_agentshadow - preload_config and gid are None")
        shadow = AgentShadow(self._topics, self._service_timings, self._agent_timings, self._update_available,
                             self._mongo_client, self._mqtt_client, self._logger, preload_config=preload_config,
                             gid=gid)
        return shadow



# http://code.activestate.com/recipes/134892/
# licensed under the PSF-licencse: https://docs.python.org/3/license.html
# the concrete license version is not provided by the source above


class _Getch:
    """Gets a single character from standard input.  Does not echo to the screen."""
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self): return self.impl()


class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()


getch = _Getch()



from tantamount.machine import Machine
from tantamount.fsm2dot import GetDotNotation

from hippodamia.agentshadow.states.active import Active
from hippodamia.agentshadow.states.inactive import Inactive
from hippodamia.agentshadow.states.onboarding import Onboarding
from hippodamia.agentshadow.states.onboarded import Onboarded
from hippodamia.agentshadow.states.stopped import Stopped
from hippodamia.agentshadow.states.archived import Archived
from hippodamia.agentshadow.states.event_ids import event_ids
from hippodamia.agentshadow.states.state_ids import state_ids

import threading


def create(update_available, timings, set_state_health, reset_max_level, logger):
    logger.info("creating state machine - start")

    logger.debug("creating state machine - creating states")
    states = {
        state_ids.ACTIVE: Active(update_available, set_state_health, reset_max_level, logger),
        state_ids.INACTIVE: Inactive(update_available, set_state_health, reset_max_level, logger),
        state_ids.ONBOARDING: Onboarding(update_available, set_state_health, reset_max_level, logger),
        state_ids.ONBOARDED: Onboarded(update_available, set_state_health, reset_max_level, logger),
        state_ids.STOPPED: Stopped(update_available, set_state_health, reset_max_level, logger),
        state_ids.ARCHIVED: Archived(update_available, set_state_health, reset_max_level, logger),
    }

    machine = Machine()

    logger.debug("creating state machine - adding states")
    machine.addstate(state_ids.ACTIVE, states[state_ids.ACTIVE])
    machine.addstate(state_ids.INACTIVE, states[state_ids.INACTIVE])
    machine.addstate(state_ids.ONBOARDING, states[state_ids.ONBOARDING])
    machine.addstate(state_ids.ONBOARDED, states[state_ids.ONBOARDED])
    machine.addstate(state_ids.STOPPED, states[state_ids.STOPPED])
    machine.addstate(state_ids.ARCHIVED, states[state_ids.ARCHIVED])

    logger.debug("creating state machine - set start states")
    machine.setstartstate(state_ids.STOPPED)

    logger.debug("creating state machine - adding transitions")
    machine.addtransition(state_ids.ONBOARDING, event_ids.TIMEOUT, state_ids.STOPPED)
    machine.addtransition(state_ids.ONBOARDING, event_ids.ONBOARDING_REQUEST, state_ids.ONBOARDING)
    machine.addtransition(state_ids.ONBOARDING, event_ids.ONBOARDING_RESPONSE, state_ids.ONBOARDED)

    machine.addtransition(state_ids.ONBOARDED, event_ids.TIMEOUT, state_ids.STOPPED)
    machine.addtransition(state_ids.ONBOARDED, event_ids.REGULAR_MESSAGE, state_ids.ACTIVE)
    machine.addtransition(state_ids.ONBOARDED, event_ids.ONBOARDING_REQUEST, state_ids.ONBOARDING)

    machine.addtransition(state_ids.ACTIVE, event_ids.REGULAR_MESSAGE, state_ids.ACTIVE)
    machine.addtransition(state_ids.ACTIVE, event_ids.TIMEOUT, state_ids.INACTIVE)
    machine.addtransition(state_ids.ACTIVE, event_ids.ONBOARDING_REQUEST, state_ids.ONBOARDING)
    machine.addtransition(state_ids.ACTIVE, event_ids.OFFBOARDING, state_ids.STOPPED)

    machine.addtransition(state_ids.INACTIVE, event_ids.TIMEOUT, state_ids.STOPPED)
    machine.addtransition(state_ids.INACTIVE, event_ids.ONBOARDING_REQUEST, state_ids.ONBOARDING)
    machine.addtransition(state_ids.INACTIVE, event_ids.OFFBOARDING, state_ids.STOPPED)
    machine.addtransition(state_ids.INACTIVE, event_ids.REGULAR_MESSAGE, state_ids.ACTIVE)

    machine.addtransition(state_ids.STOPPED, event_ids.ONBOARDING_REQUEST, state_ids.ONBOARDING)
    machine.addtransition(state_ids.STOPPED, event_ids.TIMEOUT, state_ids.ARCHIVED)

    machine.addtransition(state_ids.ARCHIVED, event_ids.ONBOARDING_REQUEST, state_ids.ONBOARDING)

    logger.debug("creating state machine - set timeout events")
    machine.addtimeoutevent(state_ids.ONBOARDED, event_ids.TIMEOUT, timings["onboarding_timeout"])
    machine.addtimeoutevent(state_ids.ONBOARDING, event_ids.TIMEOUT, timings["onboarding_timeout"])
    machine.addtimeoutevent(state_ids.ACTIVE, event_ids.TIMEOUT, timings["wait_for_message_timeout"])
    machine.addtimeoutevent(state_ids.INACTIVE, event_ids.TIMEOUT, timings["deactivation_timeout"])
    machine.addtimeoutevent(state_ids.STOPPED, event_ids.TIMEOUT, timings["archivation_timeout"])

    logger.info("creating state machine - done")
    return machine, states


def dot2file(filename):
    class NoLogger:
        def info(self, message):
            pass

        def debug(self, message):
            pass

        def warning(self, message):
            pass

        def error(self, message):
            pass

        def getChild(self, name):
            return self

    logger = NoLogger()
    updateavailable = threading.Event()
    timings = {
        "onboarding_timeout": 60,
        "wait_for_message_timeout": 180,
        "deactivation_timeout": 360,
        "heartbeat_interval": 120,
        "archivation_timeout": 86400
    }
    machine, states = create(updateavailable, timings, None, None, logger)

    gdn = GetDotNotation(machine, getStateId=(lambda x:x.name), getStateName=(lambda x:x.name),
                         getTransitionName=(lambda x:x.name))
    new_dotnotation = gdn.getdotnotation()

    try:
        with open(filename, 'r') as f:
            old_dotnotation = f.read()
    except OSError:
        old_dotnotation = ""

    if new_dotnotation != old_dotnotation:
        print("updating {} to latest version.".format(filename))
        with open(filename, "w") as f:
            f.write(new_dotnotation)



import datetime
from tantamount.astate import AState
import pelops.mylogger


class AMonitoringState(AState):
    logger = None
    update_available = None
    counter_on_exit = 0
    counter_on_entry = 0
    timestamp_on_entry = None
    timestamp_on_exit = None
    set_health = None
    health = None
    reset_max_level = None

    def __init__(self, update_available, set_health, health, reset_max_level, logger, logger_name):
        AState.__init__(self)
        self.logger = pelops.mylogger.get_child(logger, logger_name)
        self.logger.debug("__init__ ")
        self.set_health = set_health
        self.health = health
        self.update_available = update_available
        self.reset_max_level = reset_max_level

    def on_entry(self):
        self.logger.info("on_entry")
        self.counter_on_entry += 1
        self.timestamp_on_exit = None
        self.timestamp_on_entry = datetime.datetime.now()

        self.set_health(self.health)
        next = self._on_entry()

        self.update_available.set()
        if next is not None:
            self.logger.debug("on_entry - event: {}".format(next.name))
        return next

    def _on_entry(self):
        return None

    def on_exit(self):
        self.logger.info("on_exit")
        self.counter_on_exit += 1
        self.timestamp_on_exit = datetime.datetime.now()

        self._on_exit()

    def _on_exit(self):
        pass
